@@ -16,6 +16,30 @@
 
 module Gio
   class Settings
+    alias_method :initialize_raw, :initialize
+
+    def initialize(*args)
+      if args.size == 1
+        initialize_raw(args[0])
+      elsif args.size == 2
+        schema_id = args[0]
+        options = args[1]
+        path = options[:path] || nil
+        backend = options[:backend] || nil
+        if path && backend
+          initialize_new_with_backend_and_path(schema_id, backend, path)
+        elsif path
+          initialize_new_with_path(schema_id, path)
+        elsif backend
+          initialize_new_with_backend(schema_id, backend)
+        end
+      elsif args.size == 3
+        initialize_new_full(*args)
+      else
+        $stderr.puts "Arguments error for Gio::Settings#new"
+      end
+    end
+
     alias_method :set_value_raw, :set_value
     def set_value(key, value)
       schema_key = settings_schema.get_key(key)
