@@ -1,4 +1,4 @@
-# Copyright (C) 2012-2015  Ruby-GNOME2 Project Team
+# Copyright (C) 2012-2017  Ruby-GNOME2 Project Team
 #
 # This library is free software; you can redistribute it and/or
 # modify it under the terms of the GNU Lesser General Public
@@ -37,7 +37,10 @@ module GObjectIntrospection
       repository.require(namespace, @version)
       pre_load(repository, namespace)
       repository.each(namespace) do |info|
-        load_info(info)
+        load_info(info) if info.is_a?(InterfaceInfo)
+      end
+      repository.each(namespace) do |info|
+        load_info(info) unless info.is_a?(InterfaceInfo)
       end
       post_load(repository, namespace)
     end
@@ -115,9 +118,11 @@ module GObjectIntrospection
         klass = self.class.define_struct(info.size, name, @base_module,
                                          :parent => options[:parent])
       else
+        size = info.size
+        size = nil if size.zero?
         klass = self.class.define_class(info.gtype, name, @base_module,
                                         :parent => options[:parent],
-                                        :size   => info.size)
+                                        :size   => size)
       end
       load_fields(info, klass)
       load_methods(info, klass)
@@ -130,7 +135,7 @@ module GObjectIntrospection
         method_infos = info.methods.find_all do |method_info|
           method_info.is_a?(MethodInfo)
         end
-        unless methods.empty?
+        unless method_infos.empty?
           base_class = @base_module.const_get(base_class_name)
           load_method_infos(method_infos, base_class.singleton_class)
         end
@@ -230,29 +235,48 @@ module GObjectIntrospection
       field_info.name
     end
 
-    def load_field(info, i, field_info, klass)
+    def load_field(info, i, field_info, klass, options={})
       name = field_name(field_info, klass)
       flags = field_info.flags
 
-      if flags.readable?
-        if field_info.type.tag == TypeTag::BOOLEAN
-          reader_method_name = "#{name}?"
-        else
-          reader_method_name = name
-        end
-        remove_existing_method(klass, reader_method_name)
-        klass.__send__(:define_method, reader_method_name) do ||
-          info.get_field_value(self, i)
-        end
+      readable = options[:readable]
+      readable = flags.readable? if readable.nil?
+      if readable
+        reader_method_name = rubyish_field_reader_name(field_info, name)
+        load_field_reader(info, i, field_info, klass, name, reader_method_name)
       end
 
-      if flags.writable?
-        klass.__send__(:define_method, "#{name}=") do |value|
-          info.set_field_value(self, i, value)
+      writable = options[:writable]
+      writable = flags.writable? if writable.nil?
+      if writable
+        load_field_writer(info, i, field_info, klass, name, "#{name}=")
+      end
+    end
+
+    def load_field_reader(info, i, field_info, klass, name, method_name)
+      remove_existing_method(klass, method_name)
+
+      need_number_to_bool_convert = false
+      if name.start_with?("is_") and
+          field_info.type.tag == TypeTag::UINT32
+        need_number_to_bool_convert = true
+      end
+      klass.__send__(:define_method, method_name) do ||
+        value = info.get_field_value(self, i)
+        if need_number_to_bool_convert
+          value != 0
+        else
+          value
         end
       end
     end
 
+    def load_field_writer(info, i, field_info, klass, name, method_name)
+      klass.__send__(:define_method, method_name) do |value|
+        info.set_field_value(self, i, value)
+      end
+    end
+
     def load_methods(info, klass)
       grouped_methods = info.methods.group_by do |method_info|
         method_info.class
@@ -270,6 +294,10 @@ module GObjectIntrospection
           raise "TODO: #{method_info_class}"
         end
       end
+      if klass.method_defined?(:each) and
+          klass.instance_method(:each).owner == klass
+        klass.include(Enumerable)
+      end
     end
 
     def load_constructor_infos(infos, klass)
@@ -302,8 +330,6 @@ module GObjectIntrospection
     end
 
     def initialize_post(object)
-      return unless object.is_a?(GLib::Object)
-      self.class.reference_gobject(object, :sink => true)
     end
 
     def find_suitable_callable_info(infos, arguments)
@@ -408,15 +434,37 @@ module GObjectIntrospection
       if options[:prefix]
         name = name.gsub(/\A#{Regexp.escape(options[:prefix])}/, "")
       end
+      if name == "initialize"
+        name += "_raw"
+      end
+
+      n_in_args = function_info.n_in_args
+      if options[:n_in_args_offset]
+        n_in_args += options[:n_in_args_offset]
+      end
+
       return_type = function_info.return_type
       return_type_tag = return_type.tag
+      if return_type_tag == TypeTag::VOID
+        out_arg_tags = function_info.out_args.collect {|arg| arg.type.tag}
+        if out_arg_tags == [TypeTag::ARRAY, TypeTag::INT32]
+          return_type_tag = TypeTag::ARRAY
+        end
+      end
+
       case return_type_tag
       when TypeTag::BOOLEAN
         case name
+        when "equal"
+          if n_in_args == 1
+            "=="
+          else
+            name
+          end
         when /\A(?:is|get_is|can_be)_/
           "#{$POSTMATCH}?"
         when /\Aget_/
-          if function_info.n_in_args.zero?
+          if n_in_args.zero?
             if function_info.n_out_args.zero?
               "#{$POSTMATCH}?"
             else
@@ -432,10 +480,10 @@ module GObjectIntrospection
         else
           name
         end
-      when TypeTag::GLIST, TypeTag::GSLIST
+      when TypeTag::GLIST, TypeTag::GSLIST, TypeTag::ARRAY
         case name
         when /\A(?:list|get)_/
-          if function_info.n_in_args.zero?
+          if n_in_args.zero?
             $POSTMATCH
           else
             name
@@ -446,7 +494,7 @@ module GObjectIntrospection
       else
         case name
         when /\Aget_/
-          if function_info.n_in_args.zero?
+          if n_in_args.zero?
             $POSTMATCH
           else
             name
@@ -461,6 +509,21 @@ module GObjectIntrospection
       end
     end
 
+    def rubyish_field_reader_name(field_info, name)
+      case field_info.type.tag
+      when TypeTag::BOOLEAN
+        name.gsub(/\Ais_/, "") + "?"
+      when TypeTag::UINT32
+        if /\Ais_/ =~ name
+          "#{$POSTMATCH}?"
+        else
+          name
+        end
+      else
+        name
+      end
+    end
+
     def rubyish_class_name(info)
       name = info.name
       case info
@@ -490,6 +553,7 @@ module GObjectIntrospection
     def load_method_info(info, klass, method_name)
       define_method(info, klass, method_name)
       define_equal_style_setter(info, klass, method_name)
+      define_inspect(info, klass, method_name)
     end
 
     def remove_existing_method(klass, method_name)
@@ -535,6 +599,14 @@ module GObjectIntrospection
       end
     end
 
+    def define_inspect(info, klass, method_name)
+      if method_name == "to_s" and info.n_args.zero?
+        klass.__send__(:define_method, "inspect") do ||
+          super().gsub(/\>\z/) {" #{to_s}>"}
+        end
+      end
+    end
+
     def load_function_infos(infos, klass)
       infos.each do |info|
         name = rubyish_method_name(info)
