@@ -17,7 +17,10 @@
 require "test-unit"
 
 module GOfficeTestUtils
-  def omit_on_travis_ci
-    omit("This test can't be run on Travis CI") if ENV["CI"]
+  private
+  def only_goffice_version(major, minor, micro=nil)
+    unless GOffice::Version.or_later?(major, minor, micro)
+      omit("Require GOffice >= #{major}.#{minor}.#{micro}")
+    end
   end
 end
