@@ -1,5 +1,5 @@
-require 'open-uri'
-require 'fileutils'
+require "open-uri"
+require "fileutils"
 
 module PopplerTestUtils
   def ensure_dir(dir)
@@ -15,43 +15,55 @@ module PopplerTestUtils
     ensure_dir(File.join(test_dir, "fixtures"))
   end
 
+  def fixture_path(*components)
+    File.join(fixtures_dir, *components)
+  end
+
   def tmp_dir
     ensure_dir(File.join(test_dir, "tmp"))
   end
 
   def form_pdf
-    file = File.join(fixtures_dir, "form.pdf")
-    return file if File.exist?(file)
-    pdf = open("https://www.irs.gov/pub/irs-pdf/fw9.pdf").read
-    File.open(file, "wb") do |output|
-      output.print(pdf)
+    path = fixture_path("form.pdf")
+    unless File.exist?(path)
+      pdf = open("https://www.irs.gov/pub/irs-pdf/fw9.pdf").read
+      File.open(path, "wb") do |output|
+        output.print(pdf)
+      end
     end
-    file
+    path
+  end
+
+  def text_field_pdf
+    fixture_path("text-field.pdf")
   end
 
   def image_pdf
-    File.join(fixtures_dir, "image.pdf")
+    fixture_path("image.pdf")
   end
 
   def image_png
-    File.join(fixtures_dir, "image.png")
+    fixture_path("image.png")
   end
 
-  def later_version?(major, minor, micro=nil)
-    micro ||= 0
-    (Poppler::BUILD_VERSION <=> [major, minor, micro]) >= 0
+  def outline_pdf
+    fixture_path("outline.pdf")
+  end
+
+  def multiple_pages_pdf
+    fixture_path("multiple-pages.pdf")
   end
 
   def only_poppler_version(major, minor, micro=nil)
     micro ||= 0
-    unless later_version?(major, minor, micro)
+    unless Poppler::Version.or_later?(major, minor, micro)
       omit("Require Poppler >= #{major}.#{minor}.#{micro}")
     end
   end
 
   def only_old_poppler_version(major, minor, micro=nil)
     micro ||= 0
-    if later_version?(major, minor, micro)
+    if Poppler::Version.or_later?(major, minor, micro)
       omit("Require Poppler < #{major}.#{minor}.#{micro}")
     end
   end
